@@ -28,7 +28,7 @@
  *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
  *    given Attribution for the derivative code and a license granting use.
  *
- *     Copyright (C) 2004-2006 Open-Xchange, Inc.
+ *     Copyright (C) 2004-2010 Open-Xchange, Inc.
  *     Mail: info@open-xchange.com
  *
  *
@@ -54,8 +54,10 @@ import java.sql.PreparedStatement;
 import java.sql.ResultSet;
 import java.sql.SQLException;
 import java.sql.Statement;
+import java.util.ArrayList;
 import java.util.HashSet;
 import java.util.Iterator;
+import java.util.List;
 import java.util.Set;
 import java.util.UUID;
 import com.openexchange.database.DBPoolingException;
@@ -83,68 +85,6 @@ public final class MALPollDBUtility {
         super();
     }
 
-    /**
-     * Rolls-back specified connection.
-     * 
-     * @param con The connection to roll back.
-     */
-    public static void rollback(final Connection con) {
-        if (null == con) {
-            return;
-        }
-        try {
-            con.rollback();
-        } catch (final SQLException e) {
-            LOG.error(e.getMessage(), e);
-        }
-    }
-
-    /**
-     * Convenience method to set the auto-commit of a connection to <code>true</code>.
-     * 
-     * @param con The connection that should go into auto-commit mode.
-     */
-    public static void autocommit(final Connection con) {
-        if (null == con) {
-            return;
-        }
-        try {
-            con.setAutoCommit(true);
-        } catch (final SQLException e) {
-            LOG.error(e.getMessage(), e);
-        }
-    }
-
-    /**
-     * Closes the {@link ResultSet}.
-     * 
-     * @param result <code>null</code> or a {@link ResultSet} to close.
-     */
-    public static void closeSQLStuff(final ResultSet result) {
-        if (result != null) {
-            try {
-                result.close();
-            } catch (final SQLException e) {
-                LOG.error(e.getMessage(), e);
-            }
-        }
-    }
-
-    /**
-     * Closes the {@link Statement}.
-     * 
-     * @param stmt <code>null</code> or a {@link Statement} to close.
-     */
-    public static void closeSQLStuff(final Statement stmt) {
-        if (null != stmt) {
-            try {
-                stmt.close();
-            } catch (final SQLException e) {
-                LOG.error(e.getMessage(), e);
-            }
-        }
-    }
-
     /**
      * Inserts the mail IDs associated with specified hash.
      * 
@@ -154,13 +94,8 @@ public final class MALPollDBUtility {
      * @return The mail IDs associated with specified hash
      * @throws PushException If a database resource could not be acquired
      */
-    public static void insertMailIDs(final String hash, final Set<String> mailIds, final int cid) throws PushException {
-        final DatabaseService databaseService;
-        try {
-            databaseService = MALPollServiceRegistry.getServiceRegistry().getService(DatabaseService.class, true);
-        } catch (final ServiceException e) {
-            throw new PushException(e);
-        }
+    public static void insertMailIDs(final UUID hash, final Set<String> mailIds, final int cid) throws PushException {
+        final DatabaseService databaseService = getDBService();
         final Connection writableConnection;
         try {
             writableConnection = databaseService.getWritable(cid);
@@ -171,8 +106,8 @@ public final class MALPollDBUtility {
             throw PushExceptionCodes.UNEXPECTED_ERROR.create(e, e.getMessage());
         }
         try {
-            insert0(hash, mailIds, CHUNK_SIZE, writableConnection);
-
+            final StringBuilder sb = new StringBuilder(CHUNK_SIZE * 16).append(SQL_INSERT_PREFIX);
+            insert0(cid, hash, mailIds, CHUNK_SIZE, sb, writableConnection);
             writableConnection.commit(); // COMMIT
         } catch (final SQLException e) {
             rollback(writableConnection);
@@ -186,23 +121,75 @@ public final class MALPollDBUtility {
         }
     }
 
-    private static final String SQL_INSERT = "INSERT INTO malPollUid (hash, uid) VALUES (?, ?)";
+    // private static final String SQL_INSERT = "INSERT INTO malPollUid (hash, uid) VALUES (?, ?)";
 
-    private static void insert0(final String hash, final Set<String> mailIds, final int chunkSize, final Connection writableConnection) throws SQLException {
+    private static final String SQL_INSERT_PREFIX = "INSERT INTO malPollUid (cid, hash, uid) VALUES ";
+
+    private static final String SQL_INSERT_VALUES = "(?, UNHEX(REPLACE(?,'-','')), ?)";
+
+    private static void insert0(final int cid, final UUID hash, final Set<String> mailIds, final int chunkSize, final StringBuilder sb, final Connection writableConnection) throws SQLException {
         if (mailIds.isEmpty()) {
             return;
         }
+        final String uuidStr = hash.toString();
         final int isize = mailIds.size() + 1;
         final Iterator<String> iter = mailIds.iterator();
         for (int k = 1; k < isize;) {
-            final PreparedStatement stmt = writableConnection.prepareStatement(SQL_INSERT);
+            PreparedStatement stmt = null;
             try {
                 int j = k;
                 k += chunkSize;
                 final int limit = Math.min(k, isize);
+                /*
+                 * Compose statement string
+                 */
+                {
+                    sb.setLength(SQL_INSERT_PREFIX.length());
+                    final String values = SQL_INSERT_VALUES; // INSERT INTO malPollUid (hash, uid) VALUES (?, ?)
+                    sb.append(values);
+                    final String delim = ", ";
+                    for (int i = j + 1; i < limit; i++) {
+                        sb.append(delim).append(values); // , VALUES (?, ?)
+                    }
+                    stmt = writableConnection.prepareStatement(sb.toString());
+                }
+                /*
+                 * Fill values
+                 */
+                int pos = 1;
                 for (; j < limit; j++) {
-                    stmt.setString(1, hash);
-                    stmt.setString(2, iter.next());
+                    stmt.setInt(pos++, cid);
+                    stmt.setString(pos++, uuidStr);
+                    stmt.setString(pos++, iter.next());
+                }
+                stmt.executeUpdate();
+            } finally {
+                MALPollDBUtility.closeSQLStuff(stmt);
+            }
+        }
+    }
+
+    private static final String SQL_DELETE = "DELETE FROM malPollUid WHERE cid = ? AND hash = UNHEX(REPLACE(?,'-','')) AND uid = ?";
+
+    private static void deletet0(final int cid, final UUID hash, final Set<String> mailIds, final int chunkSize, final Connection writableConnection) throws SQLException {
+        if (mailIds.isEmpty()) {
+            return;
+        }
+        final String uuidStr = hash.toString();
+        final int isize = mailIds.size() + 1;
+        final Iterator<String> iter = mailIds.iterator();
+        for (int k = 1; k < isize;) {
+            final PreparedStatement stmt = writableConnection.prepareStatement(SQL_DELETE);
+            try {
+                int j = k;
+                k += chunkSize;
+                final int limit = Math.min(k, isize);
+                int pos;
+                for (; j < limit; j++) {
+                    pos = 1;
+                    stmt.setInt(pos++, cid);
+                    stmt.setString(pos++, uuidStr);
+                    stmt.setString(pos, iter.next());
                     stmt.addBatch();
                 }
                 stmt.executeBatch();
@@ -216,17 +203,13 @@ public final class MALPollDBUtility {
      * Replaces the mail IDs associated with specified hash.
      * 
      * @param hash The hash
-     * @param mailIds The new mail IDs
+     * @param newIds The new mail IDs
+     * @param delIds The deleted mail IDs
      * @param cid The context ID
      * @throws PushException If a database resource could not be acquired
      */
-    public static void replaceMailIDs(final String hash, final Set<String> mailIds, final int cid) throws PushException {
-        final DatabaseService databaseService;
-        try {
-            databaseService = MALPollServiceRegistry.getServiceRegistry().getService(DatabaseService.class, true);
-        } catch (final ServiceException e) {
-            throw new PushException(e);
-        }
+    public static void replaceMailIDs(final UUID hash, final Set<String> newIds, final Set<String> delIds, final int cid) throws PushException {
+        final DatabaseService databaseService = getDBService();
         final Connection writableConnection;
         try {
             writableConnection = databaseService.getWritable(cid);
@@ -237,25 +220,17 @@ public final class MALPollDBUtility {
             throw PushExceptionCodes.UNEXPECTED_ERROR.create(e, e.getMessage());
         }
         try {
-            PreparedStatement stmt = null;
-            try {
-                stmt = writableConnection.prepareStatement("DELETE FROM malPollUid WHERE hash = ?");
-                stmt.setString(1, hash);
-                stmt.executeUpdate();
-                MALPollDBUtility.closeSQLStuff(stmt);
+            deletet0(cid, hash, delIds, CHUNK_SIZE, writableConnection);
+            final StringBuilder sb = new StringBuilder(CHUNK_SIZE * 16).append(SQL_INSERT_PREFIX);
+            insert0(cid, hash, newIds, CHUNK_SIZE, sb, writableConnection);
 
-                insert0(hash, mailIds, CHUNK_SIZE, writableConnection);
-
-                writableConnection.commit(); // COMMIT
-            } catch (final SQLException e) {
-                rollback(writableConnection);
-                throw PushExceptionCodes.UNEXPECTED_ERROR.create(e, e.getMessage());
-            } catch (final Exception e) {
-                rollback(writableConnection);
-                throw PushExceptionCodes.UNEXPECTED_ERROR.create(e, e.getMessage());
-            } finally {
-                MALPollDBUtility.closeSQLStuff(stmt);
-            }
+            writableConnection.commit(); // COMMIT
+        } catch (final SQLException e) {
+            rollback(writableConnection);
+            throw PushExceptionCodes.UNEXPECTED_ERROR.create(e, e.getMessage());
+        } catch (final Exception e) {
+            rollback(writableConnection);
+            throw PushExceptionCodes.UNEXPECTED_ERROR.create(e, e.getMessage());
         } finally {
             autocommit(writableConnection);
             databaseService.backWritable(cid, writableConnection);
@@ -272,25 +247,7 @@ public final class MALPollDBUtility {
      * @throws PushException If a database resource could not be acquired
      */
     public static void dropMailIDs(final int cid, final int user, final int accountId, final String fullname) throws PushException {
-        /*
-         * Get hash
-         */
-        final String hash = getHash(cid, user, accountId, fullname);
-        if (null == hash) {
-            /*
-             * No hash available
-             */
-            return;
-        }
-        /*
-         * Drop everything related to hash
-         */
-        final DatabaseService databaseService;
-        try {
-            databaseService = MALPollServiceRegistry.getServiceRegistry().getService(DatabaseService.class, true);
-        } catch (final ServiceException e) {
-            throw new PushException(e);
-        }
+        final DatabaseService databaseService = getDBService();
         final Connection writableConnection;
         try {
             writableConnection = databaseService.getForUpdateTask(cid);
@@ -302,14 +259,40 @@ public final class MALPollDBUtility {
         }
         try {
             PreparedStatement stmt = null;
+            ResultSet rs = null;
             try {
-                stmt = writableConnection.prepareStatement("DELETE FROM malPollUid WHERE hash = ?");
-                stmt.setString(1, hash);
-                stmt.executeUpdate();
-                MALPollDBUtility.closeSQLStuff(stmt);
-
-                stmt = writableConnection.prepareStatement("DELETE FROM malPollHash WHERE hash = ?");
-                stmt.setString(1, hash);
+                /*
+                 * Select all UUIDs belonging to given user
+                 */
+                stmt = writableConnection.prepareStatement("SELECT hash FROM malPollHash WHERE cid = ? AND user = ?");
+                stmt.setInt(1, cid);
+                stmt.setInt(2, user);
+                rs = stmt.executeQuery();
+                final List<UUID> uuids = new ArrayList<UUID>();
+                while (rs.next()) {
+                    uuids.add(toUUID(rs.getBytes(1)));
+                }
+                rs.close();
+                stmt.close();
+                /*
+                 * Delete all entries belonging to identified UUIDs
+                 */
+                if (!uuids.isEmpty()) {
+                    stmt = writableConnection.prepareStatement("DELETE FROM malPollUid WHERE cid = ? AND hash = UNHEX(REPLACE(?,'-',''))");
+                    for (final UUID uuid : uuids) {
+                        stmt.setInt(1, cid);
+                        stmt.setString(2, uuid.toString());
+                        stmt.addBatch();
+                    }
+                    stmt.executeBatch();
+                    stmt.close();
+                }
+                /*
+                 * Delete all user data
+                 */
+                stmt = writableConnection.prepareStatement("DELETE FROM malPollHash WHERE cid = ? AND user = ?");
+                stmt.setInt(1, cid);
+                stmt.setInt(2, user);
                 stmt.executeUpdate();
                 writableConnection.commit(); // COMMIT
             } catch (final SQLException e) {
@@ -319,6 +302,7 @@ public final class MALPollDBUtility {
                 rollback(writableConnection);
                 throw PushExceptionCodes.UNEXPECTED_ERROR.create(e, e.getMessage());
             } finally {
+                MALPollDBUtility.closeSQLStuff(rs);
                 MALPollDBUtility.closeSQLStuff(stmt);
             }
         } finally {
@@ -335,13 +319,8 @@ public final class MALPollDBUtility {
      * @return The mail IDs associated with specified hash
      * @throws PushException If a database resource could not be acquired
      */
-    public static Set<String> getMailIDs(final String hash, final int cid) throws PushException {
-        final DatabaseService databaseService;
-        try {
-            databaseService = MALPollServiceRegistry.getServiceRegistry().getService(DatabaseService.class, true);
-        } catch (final ServiceException e) {
-            throw new PushException(e);
-        }
+    public static Set<String> getMailIDs(final UUID hash, final int cid) throws PushException {
+        final DatabaseService databaseService = getDBService();
         final Connection readableConnection;
         try {
             readableConnection = databaseService.getReadOnly(cid);
@@ -352,8 +331,9 @@ public final class MALPollDBUtility {
             PreparedStatement stmt = null;
             ResultSet rs = null;
             try {
-                stmt = readableConnection.prepareStatement("SELECT uid FROM malPollUid WHERE hash = ?");
-                stmt.setString(1, hash);
+                stmt = readableConnection.prepareStatement("SELECT uid FROM malPollUid WHERE cid = ? AND hash = UNHEX(REPLACE(?,'-',''))");
+                stmt.setInt(1, cid);
+                stmt.setString(2, hash.toString());
                 rs = stmt.executeQuery();
                 final Set<String> ids = new HashSet<String>();
                 while (rs.next()) {
@@ -382,13 +362,8 @@ public final class MALPollDBUtility {
      * @return The read hash or <code>null</code> if none present
      * @throws PushException If a database resource could not be acquired
      */
-    public static String getHash(final int cid, final int user, final int accountId, final String fullname) throws PushException {
-        final DatabaseService databaseService;
-        try {
-            databaseService = MALPollServiceRegistry.getServiceRegistry().getService(DatabaseService.class, true);
-        } catch (final ServiceException e) {
-            throw new PushException(e);
-        }
+    public static UUID getHash(final int cid, final int user, final int accountId, final String fullname) throws PushException {
+        final DatabaseService databaseService = getDBService();
         final Connection readableConnection;
         try {
             readableConnection = databaseService.getReadOnly(cid);
@@ -405,9 +380,12 @@ public final class MALPollDBUtility {
                 stmt.setInt(pos++, cid);
                 stmt.setInt(pos++, user);
                 stmt.setInt(pos++, accountId);
-                stmt.setString(pos++, fullname);
+                stmt.setString(pos, fullname);
                 rs = stmt.executeQuery();
-                return rs.next() ? rs.getString(1) : null;
+                if (!rs.next()) {
+                    return null;
+                }
+                return toUUID(rs.getBytes(1));
             } catch (final SQLException e) {
                 LOG.error(e.getMessage(), e);
                 return null;
@@ -430,13 +408,8 @@ public final class MALPollDBUtility {
      * @return The generated hash or <code>null</code> on failure
      * @throws PushException If a database resource could not be acquired
      */
-    public static String insertHash(final int cid, final int user, final int accountId, final String fullname) throws PushException {
-        final DatabaseService databaseService;
-        try {
-            databaseService = MALPollServiceRegistry.getServiceRegistry().getService(DatabaseService.class, true);
-        } catch (final ServiceException e) {
-            throw new PushException(e);
-        }
+    public static UUID insertHash(final int cid, final int user, final int accountId, final String fullname) throws PushException {
+        final DatabaseService databaseService = getDBService();
         final Connection writableConnection;
         try {
             writableConnection = databaseService.getWritable(cid);
@@ -447,14 +420,14 @@ public final class MALPollDBUtility {
             PreparedStatement stmt = null;
             try {
                 stmt =
-                    writableConnection.prepareStatement("INSERT INTO malPollHash (cid, user, id, fullname, hash) VALUES (?, ?, ?, ?, ?)");
+                    writableConnection.prepareStatement("INSERT INTO malPollHash (cid, user, id, fullname, hash) VALUES (?, ?, ?, ?, UNHEX(REPLACE(?,'-','')))");
                 int pos = 1;
                 stmt.setInt(pos++, cid);
                 stmt.setInt(pos++, user);
                 stmt.setInt(pos++, accountId);
                 stmt.setString(pos++, fullname);
-                final String hash = randomUUID();
-                stmt.setString(pos++, hash);
+                final UUID hash = UUID.randomUUID();
+                stmt.setString(pos, hash.toString());
                 stmt.executeUpdate();
                 return hash;
             } catch (final SQLException e) {
@@ -468,18 +441,108 @@ public final class MALPollDBUtility {
     }
 
     /**
-     * Generates a UUID using {@link UUID#randomUUID()} and removes all dashes; e.g.:<br>
-     * <i>a5aa65cb-6c7e-4089-9ce2-b107d21b9d15</i> would be <i>a5aa65cb6c7e40899ce2b107d21b9d15</i>
+     * Gets the {@link DatabaseService database service} from service registry.
      * 
-     * @return A UUID string
+     * @return The database service
+     * @throws PushException If database service is not available
      */
-    private static String randomUUID() {
-        final StringBuilder s = new StringBuilder(36).append(UUID.randomUUID());
-        s.deleteCharAt(23);
-        s.deleteCharAt(18);
-        s.deleteCharAt(13);
-        s.deleteCharAt(8);
-        return s.toString();
+    private static DatabaseService getDBService() throws PushException {
+        final DatabaseService databaseService;
+        try {
+            databaseService = MALPollServiceRegistry.getServiceRegistry().getService(DatabaseService.class, true);
+        } catch (final ServiceException e) {
+            throw new PushException(e);
+        }
+        return databaseService;
+    }
+
+    private static final int UUID_BYTE_LENGTH = 16;
+
+    /**
+     * Generates a new {@link UUID} instance from specified byte array.
+     * 
+     * @param bytes The byte array
+     * @return A new {@link UUID} instance
+     * @throws IllegalArgumentException If passed byte array is <code>null</code> or its length is not 16
+     */
+    private static UUID toUUID(final byte[] bytes) {
+        if (null == bytes) {
+            throw new IllegalArgumentException("Byte array is null.");
+        }
+        if (bytes.length != UUID_BYTE_LENGTH) {
+            throw new IllegalArgumentException("UUID must be contructed using a byte array with length 16.");
+        }
+        long msb = 0;
+        long lsb = 0;
+        for (int i = 0; i < 8; i++) {
+            msb = (msb << 8) | (bytes[i] & 0xff);
+        }
+        for (int i = 8; i < 16; i++) {
+            lsb = (lsb << 8) | (bytes[i] & 0xff);
+        }
+        return new UUID(msb, lsb);
+    }
+
+    /**
+     * Rolls-back specified connection.
+     * 
+     * @param con The connection to roll back.
+     */
+    private static void rollback(final Connection con) {
+        if (null == con) {
+            return;
+        }
+        try {
+            con.rollback();
+        } catch (final SQLException e) {
+            LOG.error(e.getMessage(), e);
+        }
+    }
+
+    /**
+     * Convenience method to set the auto-commit of a connection to <code>true</code>.
+     * 
+     * @param con The connection that should go into auto-commit mode.
+     */
+    private static void autocommit(final Connection con) {
+        if (null == con) {
+            return;
+        }
+        try {
+            con.setAutoCommit(true);
+        } catch (final SQLException e) {
+            LOG.error(e.getMessage(), e);
+        }
+    }
+
+    /**
+     * Closes the {@link ResultSet}.
+     * 
+     * @param result <code>null</code> or a {@link ResultSet} to close.
+     */
+    private static void closeSQLStuff(final ResultSet result) {
+        if (result != null) {
+            try {
+                result.close();
+            } catch (final SQLException e) {
+                LOG.error(e.getMessage(), e);
+            }
+        }
+    }
+
+    /**
+     * Closes the {@link Statement}.
+     * 
+     * @param stmt <code>null</code> or a {@link Statement} to close.
+     */
+    private static void closeSQLStuff(final Statement stmt) {
+        if (null != stmt) {
+            try {
+                stmt.close();
+            } catch (final SQLException e) {
+                LOG.error(e.getMessage(), e);
+            }
+        }
     }
 
 }
